#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include "types.h"

/* defines */

#define SCREEN_WIDTH    640
#define SCREEN_HEIGHT   400
#define TITLE           "BRESENHAM RUN-SLICED"

/* external variables */

/* None */

/* global variables */

/* None */

/* local variables */
static SDL_Surface* scr; /* the single SDL surface used */

/* local prototypes */
static void linedraw(U32 xstart, U32 ystart, S32 xend, S32 yend, COLOUR_TYPE colour);
static void DrawHorizontalRun(U32 **screenptr, S32 xadvance, S32 runlength, COLOUR_TYPE colour);
static void cls(void);
static void redraw(void);
static void vectorsup(S32 xcenter, S32 ycenter, S32 xlength, S32 ylength, COLOUR_TYPE colour);
static void DrawVerticalRun(U32 **screenptr, S32 xadvance, S32 runlength, COLOUR_TYPE colour);

int main(int argc, char *argv[])
{
    S32          width = SCREEN_WIDTH;
    S32          height = SCREEN_HEIGHT;
    S32          ColorDepth = 32;
    U8           done = 0;
    SDL_Event    event = {0};
    COLOUR_TYPE  setlinecolour;

    if(SDL_Init(SDL_INIT_EVERYTHING) < 0)
    {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }

    SDL_WM_SetCaption(TITLE, NULL);

    scr = SDL_SetVideoMode(width, height, ColorDepth, SDL_HWSURFACE | SDL_HWPALETTE);

    if(scr == NULL)
    {
        printf("Unable to set video: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }

    while(done != 1)
    {
        /* draw each of the four rectangles full of vectors */
        setlinecolour.r = 255;
        setlinecolour.g = 0;
        setlinecolour.b = 0;
        vectorsup((width / 4), (height / 4), (width / 4) , (height / 4) , setlinecolour);

        setlinecolour.r = 0;
        setlinecolour.g = 255;
        setlinecolour.b = 0;
        vectorsup((width * 3 / 4), (height / 4), (width / 4) , (height / 4) , setlinecolour);

        setlinecolour.r = 0;
        setlinecolour.g = 0;
        setlinecolour.b = 255;
        vectorsup((width / 4), (height * 3 / 4), (width / 4) , (height / 4) , setlinecolour);

        setlinecolour.r = 255;
        setlinecolour.g = 0;
        setlinecolour.b = 0;
        vectorsup((width * 3 / 4), (height * 3 / 4), (width / 4) , (height / 4) , setlinecolour);

       /* update screen */
        redraw();
        /* clear screen buffer for next write */
        cls();

        SDL_PollEvent(&event);

        /* is it time to quit ? */
        if(event.type == SDL_QUIT)
        {
            SDL_Quit();
            done = 1;
        }

        SDL_Delay(10); /* so it consumes less processing power */

    } /* while */

}

/****************************************************************/
/* Name: vectorsup                                              */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* subroutine to draw a rectangle full of vectors, of the       */
/* specified length and colour, around the specified rectangle  */
/* center.                                                      */
/*                                                              */
/****************************************************************/
static void vectorsup(S32 xcenter, S32 ycenter, S32 xlength, S32 ylength, COLOUR_TYPE colour)
{
    S32 workingx;
    S32 workingy;

    /* lines from the center to top of rectangle */
      workingx = xcenter - xlength;
      workingy = ycenter - ylength;
      for( ; workingx < (xcenter + xlength); workingx++)
      {
          linedraw(xcenter, ycenter, workingx, workingy, colour);
      }

    /* lines from the center to right of rectangle */
    workingx = xcenter + xlength - 1;
    workingy = ycenter - ylength;
    for( ; workingy < (ycenter + ylength); workingy++)
    {
        linedraw(xcenter, ycenter, workingx, workingy, colour);
    }

    /* lines from the center to bottom of rectangle */
    workingx = xcenter + xlength - 1;
    workingy = ycenter + ylength - 1;
    for( ; workingx >= (xcenter - xlength); workingx--)
    {
        linedraw(xcenter, ycenter, workingx, workingy, colour);
    }

    /* lines from the center to left of rectangle */
    workingx = xcenter - xlength;
    workingy = ycenter + ylength - 1;
    for( ; workingy >= (ycenter - ylength); workingy--)
    {
        linedraw(xcenter, ycenter, workingx, workingy, colour);
    }

}

/****************************************************************/
/* Name: cls                                                    */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* Clears the screen to black                                   */
/*                                                              */
/****************************************************************/
static void cls(void)
{
  SDL_FillRect(scr, NULL, 0);
}

/****************************************************************/
/* Name: redraw                                                 */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* Updates the screen.  Has to be called to view new pixels,    */
/* but use only after drawing the whole screen because it's     */
/* slow.                                                        */
/*                                                              */
/****************************************************************/
static void redraw(void)
{
  SDL_UpdateRect(scr, 0, 0, 0, 0);
}

/****************************************************************/
/* Name: linedraw                                               */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draw a line between the specified endpoints in colour colour */
/*                                                              */
/****************************************************************/
static void linedraw(U32 xstart, U32 ystart, S32 xend, S32 yend, COLOUR_TYPE colour)
{
    S32 temp;
    S32 xdelta;
    S32 ydelta;
    S32 xadvance;
    S32 i;
    U32 colorSDL;
    U32 *bufp;
    S32 wholestep;
    S32 adjup;
    S32 adjdown;
    S32 errorterm;
    S32 InitialPixelCount;
    S32 FinalPixelCount;
    S32 runlength;

    /* setup pixel colour */
    colorSDL = SDL_MapRGB(scr->format, colour.r, colour.g, colour.b);

    /* we'll always draw top to bottom, to reduce the number of cases we have */
    /* to handle, and to make lines between the same endpoints draw the same pixels */
    if(ystart > yend)
    {
        temp = ystart;
        ystart = yend;
        yend = temp;
        temp = xstart;
        xstart = xend;
        xend = temp;
    }

    /* put screen buffer pointer at the required x & y coord's*/
    bufp = (U32*)(scr->pixels + ((ystart * 4) * (scr->pitch / 4)) + (xstart * 4));

    /* figure out whether we're going left or right, and how far we're */
    /* going horizontally */
    xdelta = xend - xstart;

    if(xdelta < 0)
    {
        xadvance = -1;
        xdelta = -xdelta;
    }
    else
    {
        xadvance = 1;
    }

    /* figure out how far we're going vertically */
    ydelta = yend - ystart;

    /* special case horizontal, vertical and diagonal lines, for speed */
    /* and to avoid nasty boundary conditions and divide by zero */
    if(xdelta == 0)
    {
        /* vertical lines */
        for(i = 0 ; i <= ydelta ; i++)
        {
            /* draw a pixel of colour colorSDL at coord's x & y */
            *bufp = colorSDL;
            /* put screen buffer pointer at the required x & y coord's*/
            bufp += scr->pitch / 4;
        }
        return;
    }

    if(ydelta == 0)
    {
        /* horizontal line */
        for(i = 0; i <= xdelta; i++)
        {
            /* draw a pixel of colour colorSDL at coord's x & y */
            *bufp = colorSDL;
            /* put screen buffer pointer at the required x & y coord's*/
            bufp += xadvance;
        }
        return;
    }

    if(xdelta == ydelta)
    {
        /* diagonal line */
        for(i = 0 ; i <= xdelta ; i++)
        {
            /* draw a pixel of colour colorSDL at coord's x & y */
            *bufp = colorSDL;
            /* put screen buffer pointer at the required x & y coord's*/
            bufp += (scr->pitch / 4) + xadvance;
        }
        return;
    }

    /* determine whether the line is X or Y major, and handle accordingly */
    if(xdelta >= ydelta)
    {
        /* x major line */
        /* minimum number of pixels in a run in this line */
        wholestep = xdelta / ydelta;

        /* error term adjust each time y steps by 1 used to tell when one */
        /* extra pixel should be drawn as part of a run. to account for */
        /* fractional steps along the x axis per 1-pixel steps along y */
        adjup = (xdelta % ydelta) * 2;

        /* error term adjust when the error term turns over, used to factor */
        /* out the x step made at that time */
        adjdown = ydelta * 2;

        /* initial error term reflects an initial step of 0.5 along the y axis */
        errorterm = (xdelta % ydelta) - (ydelta * 2);

        /* the initial and last runs are partial, because y advances only 0.5 */
        /* for these runs rather than 1. divide one full run plus the initial */
        /* pixel, between the initial and last runs */
        InitialPixelCount = (wholestep / 2) + 1;
        FinalPixelCount = InitialPixelCount;

        /* if the basic run length is even and theres no fractional advance */
        /* we have one pixel that could go to either the initial or last partial */
        /* run, which we'll arbitrarily allocate to the last run */
        if((adjup == 0) && ((wholestep & 0x01) == 0))
        {
            InitialPixelCount--;
        }

        /* if there're an odd number of pixels per run, we have 1 pixel that */
        /* cant be allocated to either the initial or last partial run, so we'll */
        /* add 0.5 to error term so this pixel will be handled by the normal */
        /* full-run loop */
        if((wholestep & 0x01) != 0)
        {
            errorterm += ydelta;
        }

        /* draw the first partial run of pixels */
        DrawHorizontalRun(&bufp, xadvance, InitialPixelCount, colour);

        /* draw all full runs */
        for(i = 0; i < (ydelta-1);i++)
        {
            runlength = wholestep; /* run is at least this long */

            /* adavance the error term and add an extra pixel if the error */
            /* term so indicates */
            errorterm += adjup;

            if(errorterm > 0)
            {
                runlength++;
                errorterm -= adjdown;    /* reset the error term */
            }

            /* draw this scan line's run */
            DrawHorizontalRun(&bufp, xadvance, runlength, colour);
        }

        /* draw the final run of pixels */
        DrawHorizontalRun(&bufp, xadvance, FinalPixelCount, colour);
        return;
    }
    else
    {
        /* y major line */
        /* minimum number of pixels in a run in this line */
        wholestep = ydelta / xdelta;

        /* error term adjust each time x steps by 1 used to tell when one */
        /* extra pixel should be drawn as part of a run. to account for */
        /* fractional steps along the y axis per 1-pixel steps along x */
        adjup = (ydelta % xdelta) * 2;

        /* error term adjust when the error term turns over, used to factor */
        /* out the y step made at that time */
        adjdown = xdelta * 2;

        /* initial error term reflects an initial step of 0.5 along the x axis */
        errorterm = (ydelta % xdelta) - (xdelta * 2);

        /* the initial and last runs are partial, because x advances only 0.5 */
        /* for these runs rather than 1. divide one full run plus the initial */
        /* pixel, between the initial and last runs */
        InitialPixelCount = (wholestep / 2) + 1;
        FinalPixelCount = InitialPixelCount;

        /* if the basic run length is even and theres no fractional advance */
        /* we have one pixel that could go to either the initial or last partial */
        /* run, which we'll arbitrarily allocate to the last run */
        if((adjup == 0) && ((wholestep & 0x01) == 0))
        {
            InitialPixelCount--;
        }

        /* if there're an odd number of pixels per run, we have 1 pixel that */
        /* cant be allocated to either the initial or last partial run, so we'll */
        /* add 0.5 to error term so this pixel will be handled by the normal */
        /* full-run loop */
        if((wholestep & 0x01) != 0)
        {
            errorterm += xdelta;
        }

        /* draw the first partial run of pixels */
        DrawVerticalRun(&bufp, xadvance, InitialPixelCount, colour);

        /* draw all full runs */
        for(i = 0; i < (xdelta-1);i++)
        {
            runlength = wholestep; /* run is at least this long */

            /* adavance the error term and add an extra pixel if the error */
            /* term so indicates */
            errorterm += adjup;

            if(errorterm > 0)
            {
                runlength++;
                errorterm -= adjdown; /* reet the error term */
            }

            /* draw this scan line's run */
            DrawVerticalRun(&bufp, xadvance, runlength, colour);
        }

        /* draw the final run of pixels */
        DrawVerticalRun(&bufp, xadvance, FinalPixelCount, colour);
        return;
    }

}

/****************************************************************/
/* Name: DrawHorizontalRun                                      */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draws a horizontal run of pixels then advances the bitmap    */
/* pointer to the first pixel of the next run                   */
/*                                                              */
/****************************************************************/
static void DrawHorizontalRun(U32 **screenptr, S32 xadvance, S32 runlength, COLOUR_TYPE colour)
{
    S32 i;
    U32 *workingscreenptr = *screenptr;
    U32 colorSDL;

    /* setup pixel colour */
    colorSDL = SDL_MapRGB(scr->format, colour.r, colour.g, colour.b);

    for(i = 0; i < runlength; i++)
    {
        *workingscreenptr = colorSDL;
        workingscreenptr += xadvance;
    }

    /* advance to the next scan line */
    workingscreenptr += (scr->pitch / 4);
    *screenptr = workingscreenptr;
}

/****************************************************************/
/* Name: DrawVerticalRun                                        */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draws a vertical run of pixels then advances the bitmap      */
/* pointer to the first pixel of the next run                   */
/*                                                              */
/****************************************************************/
static void DrawVerticalRun(U32 **screenptr, S32 xadvance, S32 runlength, COLOUR_TYPE colour)
{
    S32 i;
    U32 *workingscreenptr = *screenptr;
    U32 colorSDL;

    /* setup pixel colour */
    colorSDL = SDL_MapRGB(scr->format, colour.r, colour.g, colour.b);

    for(i = 0; i < runlength; i++)
    {
        *workingscreenptr = colorSDL;
        workingscreenptr += (scr->pitch / 4);
    }

    /* advance to the next column */
    workingscreenptr += xadvance;
    *screenptr = workingscreenptr;
}
